# Generated by Django 3.2.16 on 2023-01-31 09:25

import localized_fields.fields.field
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_analytics", "0006_rename_form_slug_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="analyticstable",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
        migrations.AddField(
            model_name="historicalanalyticstable",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
    ]
