# Generated by Django 4.2.9 on 2024-01-23 14:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_workflow", "0031_simple_history"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflow",
            name="start_tasks",
            field=models.ManyToManyField(
                help_text="Starting task(s) of the workflow.",
                related_name="+",
                to="caluma_workflow.task",
            ),
        ),
    ]
