# Generated by Django 4.2.9 on 2024-01-23 14:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_form", "0046_file_answer_reverse_keys"),
    ]

    operations = [
        migrations.AlterField(
            model_name="answer",
            name="documents",
            field=models.ManyToManyField(
                related_name="+",
                through="caluma_form.AnswerDocument",
                to="caluma_form.document",
            ),
        ),
    ]
